/* soundmosaic, Copyright (c) 2001-2007 Steven Hazel <sah@thalassocracy.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */
#include "config.h"

#include <unistd.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <assert.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "main.h"
#include "http.h"
#include "io_buffer.h"
#include "client.h"

int writeall(int fd, const void *buf, size_t count)
{
  const char *cbuf = (char *)buf;
  int bw;
  unsigned int wlen = 0;

  while (wlen < count) {
    bw = write(fd, &(cbuf[wlen]), count - wlen);
    if (bw < 1) {
      return bw;
    }
    wlen += bw;
  }

  return wlen;
}


/* if the return value of this function is != -1 and < count, it's
   because we've encountered an EOF */
int readall(int fd, void *buf, size_t count) {
  char *cbuf = (char *)buf;
  int br;
  unsigned int rlen = 0;

  while (rlen < count) {
    br = read(fd, &(cbuf[rlen]), count - rlen);
    if (br == 0) {
      return rlen;
    } else if (br < 0) {
      return br;
    }
    rlen += br;
  }

  return rlen;
}


/* just open a tcp connection */
int open_connection (int *sock, char *hostaddress, int port)
{
  struct in_addr addr;
  int connected_socket, connected;
  struct sockaddr_in address;
  struct hostent *host;

  if ((port < 1) || (port > 65535)) {
    fprintf(stderr, "invalid port number %d\n", port);
    return FALSE;
  }

  addr.s_addr = inet_addr(hostaddress);
  if (addr.s_addr == -1) {
    host = gethostbyname(hostaddress);
    if (host == NULL) {
      fprintf(stderr, "invalid hostname\n");
      return FALSE;
    }
    addr.s_addr = ((struct in_addr *) *host->h_addr_list)->s_addr;
  }

  memset((char *) &address, 0, sizeof(address));
  address.sin_family = AF_INET;
  address.sin_port = htons(port);
  address.sin_addr.s_addr = addr.s_addr;

  connected_socket = socket(AF_INET, SOCK_STREAM, 0);

  connected = connect(connected_socket, (struct sockaddr *) &address,
                      sizeof(address));
  if (connected < 0) {
    perror("could not connect to master");
    return FALSE;
  }

  *sock = connected_socket;

  return TRUE;
}


int get_target_file(char *hostaddress, int port, FILE *temp_file)
{
  int i;
  int status;
  int sock;
  int content_length = -1;
  int total;
  char *headers;
  int count;
  int header_end_pos;
  io_buffer in_buf;
  io_buffer out_buf;
  http_request request;
  const char *request_headers =
    "GET / HTTP/1.0\x0D\x0A"
    "User-Agent: soundmosaic/" VERSION "\x0D\x0A"
    "Accept: */*\x0D\x0A"
    "Connection: Close\x0D\x0A\x0D\x0A";

  if (temp_file == NULL) {
    fprintf(stderr, "couldn't get target file: destination stream is null\n");
    return FALSE;
  }

  io_buffer_init(&in_buf, 64 * 1024);
  io_buffer_init(&out_buf, 64 * 1024);

  status = open_connection(&sock, hostaddress, port);
  if (!status) {
    return FALSE;
  }

  io_buffer_write(&out_buf, request_headers, strlen(request_headers));
  status = io_buffer_send_all(&out_buf, sock);
  if( status <= 0 ) {
    fprintf(stderr, "write failed with %d while sending HTTP request\n");
    return FALSE;
  }

  do {
    if (io_buffer_failed(&in_buf)) {
      fprintf(stderr, "buffer overflow before HTTP header end found\n");
      return FALSE;
    }

    status = io_buffer_recv_all(&in_buf, sock);
    //printf("read %d bytes\n", status);
    if (status <= 0) {
      fprintf(stderr,
              "read failed with error %d"
              " before HTTP header end found\n", status);
      return FALSE;
    }

    header_end_pos = http_find_header_end(&in_buf, 1);
  } while (header_end_pos < 0);

  headers = malloc(header_end_pos + 1);
  if (headers == NULL) {
    fprintf(stderr, "couldn't allocate %d bytes for HTTP headers\n",
            header_end_pos + 1);
    return FALSE;
  }

  io_buffer_read(&in_buf,
                 headers,
                 header_end_pos);
  headers[header_end_pos] = 0;

  http_parse_headers(headers, &request);

  free(headers);

  for (i = 0; i < request.num_headers; i++) {
    if (strcasecmp(request.keys[i], "content-length") == 0) {
      content_length = atoi(request.values[i]);
    }
  }

  if (request.type != HTTP_OK) {
    fprintf(stderr, "master response doesn't look like an OK\n");
  }

  if (content_length < 1) {
    fprintf(stderr, "invalid Content-Length %d from master\n", content_length);
    return FALSE;
  }

  printf("receiving target file from master\n");
  total = 0;
  do {
    count = io_buffer_bytes_used(&in_buf);
    if (total + count > content_length) {
      count = content_length - total;
    }
    status = io_buffer_fwrite(&in_buf, temp_file, count);
    if (status != count) {
      if (status < 0) {
        perror("error writing file");
      }
      fprintf(stderr, "failed writing file, closing connection\n");
      close(sock);
      return FALSE;
    }

    total = ftell(temp_file);

    //fprintf(stderr, "wrote %d of %d bytes to temp file\n",
    //        total, content_length);

    assert(total <= content_length);

    if (total < content_length) {
      status = io_buffer_recv_all(&in_buf, sock);
      if (status <= 0) {
        if (status < 0) {
          perror("unexpected socket error");
        }
        fprintf(stderr, "failed receiving data, closing connection\n");
        close(sock);
        return FALSE;
      }
    }
  } while (total < content_length);
  printf("finished receiving target file from master\n");

  fseek(temp_file, 0, SEEK_SET );

  return TRUE;
}


int send_output_file(char *hostaddress, int port, FILE *out_stream)
{
  char buf[128];
  int bytes_read;
  int status;
  int sock;
  int length;
  const char *end_headers = "\x0D\x0A";
  const char *request_headers =
    "POST / HTTP/1.0\x0D\x0A"
    "User-Agent: soundmosaic/" VERSION "\x0D\x0A"
    "Accept: */*\x0D\x0A"
    "Connection: Close\x0D\x0A";
  char content_length_header[80];

  if (out_stream == NULL) {
    fprintf(stderr, "couldn't send output file: source stream is null\n");
    return FALSE;
  }

  status = open_connection(&sock, hostaddress, port);
  if (!status) {
    return FALSE;
  }

  status = writeall(sock, request_headers, strlen(request_headers));
  if (status != strlen(request_headers)) {
    fprintf(stderr, "connection failed sending output to master\n");
    return FALSE;
  }

  fseek(out_stream, 0, SEEK_END);
  length = ftell(out_stream);
  snprintf(content_length_header, 80,
           "Content-Length: %d\x0D\x0A", length);
  status = writeall(sock, content_length_header,
                    strlen(content_length_header));
  if (status != strlen(content_length_header)) {
    fprintf(stderr, "connection failed sending output to master\n");
    return FALSE;
  }

  status = writeall(sock, end_headers, strlen(end_headers));
  if (status != strlen(end_headers)) {
    fprintf(stderr, "connection failed sending output to master\n");
    return FALSE;
  }

  fseek(out_stream, 0, SEEK_SET);

  do {
    bytes_read = fread(buf, 1, 128, out_stream);
    if (bytes_read == 0) {
      printf("got read 0 from file\n");
    } else {
      //printf("read %d bytes from file\n", bytes_read);

      status = writeall(sock, buf, bytes_read);
      if (status != bytes_read) {
        fprintf(stderr, "connection failed sending output to master\n");
        return FALSE;
      }
    }
  } while (bytes_read == 128);


  status = writeall(sock, end_headers, strlen(end_headers));
  if (status != strlen(end_headers)) {
    fprintf(stderr, "connection failed sending output to master\n");
    return FALSE;
  }

  close(sock);

  return TRUE;
}
